@@ -2,7 +2,7 @@
 DSLIBRARY - run your data science code anywhere.
 
 """
-__version__ = "0.0.3"
+__version__ = "0.0.4"
 
 # environment variables
 from .front import ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, ENV_DSLIBRARY_TOKEN
@@ -18,7 +18,7 @@ except ImportError:
     mmlibrary = None
 
 
-def instance():
+def instance(env: dict=None):
     """
     Get an instance of mmlibrary methods.  Configuration is through environment variables.
 
@@ -39,7 +39,9 @@ def instance():
     JSON (see ModelRunner).
       DSLIBRARY_SPEC={...}
 
-    :returns:  An instance of the MMLibrary() base class.
+    :param env:  Instead of using os.environ, environment values can be specified here.
+
+    :returns:    An instance of the MMLibrary() base class.
     """
     from .transport.to_local import DSLibraryLocal
     from .transport.to_rest import DSLibraryViaREST
@@ -49,12 +51,14 @@ def instance():
     import importlib
     import json
     import os
-    spec = json.loads(os.environ.get(ENV_DSLIBRARY_SPEC) or "{}")
-    target = os.environ.get(ENV_DSLIBRARY_TARGET) or ("mmlibrary" if mmlibrary else "local")
+    if env is None:
+        env = os.environ
+    spec = json.loads(env.get(ENV_DSLIBRARY_SPEC) or "{}")
+    target = env.get(ENV_DSLIBRARY_TARGET) or ("mmlibrary" if mmlibrary else "local")
     target_parts = target.split(":")
     # point to a REST API
     if target_parts[0] in ("http", "https"):
-        token = os.environ.get(ENV_DSLIBRARY_TOKEN)
+        token = env.get(ENV_DSLIBRARY_TOKEN)
         return DSLibraryViaREST(url=target, token=token)
     # gather arguments
     args = []
