@@ -15,7 +15,7 @@ import typing
 from collections import namedtuple
 import yaml
 
-import dslibrary.utils.connect
+from .utils.connect import connect_to_database
 from .utils.dbconn import sql_enquote_id
 from .utils.file_utils import write_stream_with_read_on_close
 
@@ -60,6 +60,7 @@ class DSLibrary(object):
         self._run_id = self._spec.get("run_id") or ""
         self._mlflow_opts = self._spec.get("mlflow") or {}
         self._mlflow_all = self._mlflow_opts.get("all")
+        self._mlflow_metrics = self._mlflow_opts.get("metrics") or self._mlflow_all
         if self._mlflow_opts and not mlflow:
             raise DSLibraryException("MLFlow was selected for output but the mlflow package is not installed")
         self._mlflow_run = None
@@ -247,7 +248,7 @@ class DSLibrary(object):
         kwargs.update(resource_spec)
         # connect to data
         uri = kwargs.pop("uri", None)
-        return dslibrary.utils.connect.connect_to_database(path=uri, for_write=for_write, **kwargs)
+        return connect_to_database(uri=uri, for_write=for_write, **kwargs)
 
     def get_nosql_connection(self, resource_name: str, **kwargs):
         """
@@ -303,8 +304,7 @@ class DSLibrary(object):
         """
         Save a piece of trackable information about this run.
         """
-        use_mlflow = self._mlflow_opts.get("metrics") or self._mlflow_all
-        if use_mlflow:
+        if self._mlflow_metrics:
             if mlflow:
                 mlflow.log_metric(metric_name, metric_value, step)
             return
@@ -403,13 +403,13 @@ class DSLibrary(object):
         for f in os.listdir(local_dir):
             self.log_artifact(os.path.join(local_dir, f), os.path.join(artifact_path, f))
 
-    def iter_scoring_requests(self, callback=None):
+    def next_scoring_request(self, timeout: float=None) -> (dict, None):
         """
-        Iterates through multiple requests.  Each iterated value is {} defining the request.
+        Use this method to iterate through multiple requests.  Returns a new {} on each request, or None if 'timeout'
+        expired before any new requests were received.
 
-        :param callback:       Called while waiting for requests, and can raise an exception to stop the iteration.
+        :param timeout:    How long to wait for a new request.
         """
-        return []
 
     def scoring_response(self, score) -> None:
         """
