@@ -10,6 +10,7 @@ import sys
 
 from pexpect import split_command_line
 
+import dslibrary
 from dslibrary import ENV_DSLIBRARY_TOKEN
 from dslibrary.front import DSLibraryException, ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, METRICS_ALIAS
 from dslibrary.metadata import Metadata
@@ -19,7 +20,7 @@ class ModelRunner(object):
     """
     Each instance sets up for one run of one model.
     """
-    def __init__(self, uri: str=None, user: str=None, run_id: str=None, entry_point: str = "main", mlflow: dict=None):
+    def __init__(self, uri: str=None, user: str=None, run_id: str=None, entry_point: str = "main", mlflow: (bool, dict)=None):
         """
         The constructor provides some context which is optional when directly invoking local code, and required
         when invoking code remotely.
@@ -34,25 +35,22 @@ class ModelRunner(object):
         self.user = user or ""
         self.run_id = run_id or ""
         self.entry_point = entry_point
-        # TODO document & validate these
         self.mlflow = mlflow if isinstance(mlflow, dict) else {"all": True} if mlflow is True else {}
         self.inputs = {}
         self.outputs = {}
         self.parameters = {}
 
-    def get_metadata(self):
-        """
-        Obtain metadata for the selected model.
-        """
-        # TODO code me ... depends quite a bit on where the model is and how it is packaged, not possible without more context
-
     def _to_absolute(self, uri: str):
         """
-        Convert local paths to absolute paths except when they match a certain pattern.
+        Convert local paths to absolute paths except when they match a certain pattern.  "~" signifies the
+        model's root path and can be used to access built-in model data.
         """
-        # TODO test this (and make sure it seems reasonable to use '~')
+        # TODO test this
+        # blank > means no change to supplied name
+        if not uri:
+            return ""
+        # no changes to URIs
         if ":" in uri:
-            # no changes to URIs
             return uri
         if uri.startswith("~/"):
             # specifically point to files within the sandbox
@@ -63,7 +61,7 @@ class ModelRunner(object):
         # convert to absolute path
         return os.path.abspath(uri)
 
-    def set_input(self, name: str, uri: str, **kwargs):
+    def set_input(self, name: str, uri: str="", **kwargs):
         """
         Specify where data will come from for a particular named input.
         :param name:    Name of input.
@@ -75,7 +73,7 @@ class ModelRunner(object):
         self.inputs[name] = {"uri": self._to_absolute(uri), **kwargs}
         return self
 
-    def set_output(self, name: str, uri: str, **kwargs):
+    def set_output(self, name: str, uri: str="", **kwargs):
         """
         Specify where data should go for a particular named output.
 
@@ -116,7 +114,6 @@ class ModelRunner(object):
         """
         if mlflow is not None:
             self.mlflow["metrics"] = bool(mlflow)
-        # TODO shouldn't the input and the output be separate?
         self.inputs[METRICS_ALIAS] = {"uri": uri, **kwargs}
         self.outputs[METRICS_ALIAS] = {"uri": uri, **kwargs}
         return self
@@ -137,27 +134,46 @@ class ModelRunner(object):
 
     def run_method(self, method: typing.Callable, path: str=None):
         """
-        As long as the target 'method' does not uses subprocesses or threads, this method can be used to invoke it.
+        As long as 'method()' uses dslibrary.instance(), and all calls through this method are single threaded, you can
+        use this method, and it is very efficient.
         """
-        # store specification
-        orig_env = {ENV_DSLIBRARY_TARGET: os.environ.get(ENV_DSLIBRARY_TARGET), ENV_DSLIBRARY_SPEC: os.environ.get(ENV_DSLIBRARY_SPEC)}
-        os.environ[ENV_DSLIBRARY_TARGET] = f"local:{path or '.'}"
-        os.environ[ENV_DSLIBRARY_SPEC] = json.dumps(self._generate_spec())
         # map parameters to method signature
         call_kwargs = {}
         sig = inspect.signature(method)
-        for p_name in sig.parameters:
-            if p_name in self.parameters:
-                param = sig.parameters[p_name]
-                value = self.parameters[p_name]
-                if param.annotation:
+        sig_kwargs = None
+        remaining_params = dict(self.parameters)
+        for p_name, param in sig.parameters.items():
+            if param.kind == inspect.Parameter.VAR_KEYWORD:
+                sig_kwargs = True
+                continue
+            if p_name in remaining_params:
+                value = remaining_params.pop(p_name)
+                if param.annotation and param.annotation is not inspect.Parameter.empty:
                     value = param.annotation(value)
                 call_kwargs[p_name] = value
-        # the method makes calls to dslibrary, following rules set above
-        method(**call_kwargs)
-        # restore environment
-        for k, v in orig_env.items():
-            os.environ[k] = v or ""
+        if sig_kwargs:
+            call_kwargs.update(remaining_params)
+        run_env = self.prepare_run_env(local_path=path)
+        # as long as the method lets us send a custom dslibrary instance we can do that
+        if "dsl" in sig.parameters or "dslibrary" in sig.parameters:
+            dsl = dslibrary.instance(run_env)
+            call_kwargs["dsl" if "dsl" in sig.parameters else "dslibrary"] = dsl
+            method(**call_kwargs)
+        else:
+            # otherwise we assume method() calls dslibrary.instance()
+            # - this approach is not thread safe
+            # TODO is this approach even worth supporting???
+            orig_env = {ENV_DSLIBRARY_TARGET: os.environ.get(ENV_DSLIBRARY_TARGET), ENV_DSLIBRARY_SPEC: os.environ.get(ENV_DSLIBRARY_SPEC)}
+            try:
+                # store specification
+                for k, v in run_env.items():
+                    os.environ[k] = v or ""
+                # the method makes calls to dslibrary, following rules set above
+                method(**call_kwargs)
+            finally:
+                # restore environment
+                for k, v in orig_env.items():
+                    os.environ[k] = v or ""
 
     def run_local(self, path: str, extra_env: dict=None):
         """
@@ -185,8 +201,7 @@ class ModelRunner(object):
             cmd = self.infer_command_from_path(path)
         # work out environment variables
         env = {
-            ENV_DSLIBRARY_TARGET: f"local:{folder}",
-            ENV_DSLIBRARY_SPEC: json.dumps(self._generate_spec()),
+            **self.prepare_run_env(local_path=folder),
             **(extra_env or {})
         }
         # call the model
@@ -197,6 +212,8 @@ class ModelRunner(object):
     def infer_command_from_path(path: str):
         """
         When we only know the name of a source file we have to work out how to execute it.
+
+        Used by run_local().
         """
         _, fn = os.path.split(path)
         _, f_ext = os.path.splitext(fn)
@@ -212,20 +229,45 @@ class ModelRunner(object):
             raise DSLibraryException(f"unrecognized executable extension: {f_ext} in {fn}")
         return cmd
 
-    def prepare_remote_run(self, local_rest_url: str, rest_access_token: str=None) -> dict:
+    def prepare_run_env(self, rest_url: str=None, shared_volume: str=None, local_path: str=None, rest_access_token: str=None) -> dict:
         """
-        Generically prepare for a remote run, assuming the remote will use make REST calls to access its data.  Returns
-        the environment variables that must be set when the remote executes.  These can go in a k8s job spec, for
-        instance.
-        :param local_rest_url:  URL of REST service.
-        :param rest_access_token: Access token to secure communications.
+        Prepare environment variables for a run which encapsulate all of the settings generated by calls to this
+        class.  Data science code running anywhere, with these environment variables set, can use dslibrary to read
+        and write all of its data without having to know any details about data location, credentials or format.
+
+        Three approaches are provided:
+          * rest_url and rest_access_token - fill this in to have the target code use a REST service for data access.
+          * shared_volume - fill this in to use a shared volume (i.e. a sidecar) for data access.
+          * local_path - this option causes all inputs and outputs to default to files in the indicated folder, and
+              there is no delegation of data access, it is all performed by the process running the target code.
+
+        In a high security scenario which isolates the data science code from the credentials and other data source
+        details:
+          * Send the ENV_DSLIBRARY_SPEC value to a secure service (a REST service indicated by rest_uri, or a sidecar
+            sharing the volume 'shared_volume' with the target code).
+          * Send the other environment variables to the target code.
+          * The target code calls dslibrary methods, which communicate with the data service, and the data service
+            performs all reads and writes based on the environment data it has been sent.
+
+        The Kubernetes sidecar approach involves running a container alongside your target code's container, and having
+        them both mount the same shared, ephemeral volume.  The appropriate environment variables are set for each, and
+        voila, the target code is able to perform all its data access without having exposed any sensitive information.
+
+        :param rest_url:            URL of REST service.
+        :param shared_volume:       Path to shared volume.
+        :param local_path:          Local path.
+        :param rest_access_token:   Access token to secure communications.
         :return:  Environment variables to set, as a {}.
         """
         env = {}
-        if local_rest_url:
-            env[ENV_DSLIBRARY_TARGET] = local_rest_url
-        if rest_access_token:
-            env[ENV_DSLIBRARY_TOKEN] = rest_access_token
+        if shared_volume:
+            env[ENV_DSLIBRARY_TARGET] = f"volume:{shared_volume}"
+        elif rest_url:
+            env[ENV_DSLIBRARY_TARGET] = rest_url
+            if rest_access_token:
+                env[ENV_DSLIBRARY_TOKEN] = rest_access_token
+        elif local_path:
+            env[ENV_DSLIBRARY_TARGET] = f"local:{local_path}"
         env[ENV_DSLIBRARY_SPEC] = json.dumps(self._generate_spec())
         return env
 
@@ -238,3 +280,4 @@ class ModelRunner(object):
         variable.
         """
         # TODO code me
+        # TODO I don't see a way to have MLFlow deliver extra environment variables!
