@@ -55,7 +55,7 @@ class DSLibraryViaMMLibrary(DSLibrary):
         method = self._find_method("get_parameter", "get_argument", "getArgument")
         try:
             return method(parameter_name)
-        except KeyError:
+        except (ValueError, KeyError):
             pass
         return default
 
@@ -75,6 +75,14 @@ class DSLibraryViaMMLibrary(DSLibrary):
                 raise FileNotFoundError(f"not found: {path}")
         raise DSLibraryException(f"Unsupported mode: {mode}")
 
+    def log_metric(self, metric_name: str, metric_value: float, step: int=0) -> None:
+        """
+        MMlibrary has its own KPI logging method.
+        """
+        if self._mlflow_metrics:
+            return super(DSLibraryViaMMLibrary, self).log_metric(metric_name, metric_value, step)
+        self._mm.save_kpi(metric_name, metric_value)
+
     def open_run_data(self, filename: str, mode: str='rb') -> io.IOBase:
         """
         The method provided by mmlibrary to save 'run data' only stores a single chunk of binary data.  We
